from dataclasses import fields as get_fields_tuple

from fmtr.tools.tools import EMPTY


def get_fields(cls, **filters) -> dict:
    """

    Get a dictionary of fields from a dataclass.

    """
    fields = {}
    for field in get_fields_tuple(cls):
        if all([getattr(field, key) is value for key, value in filters.items()]):
            fields[field.name] = field

    return fields


def get_metadata(cls, **filters) -> dict:
    """

    Get a dictionary of fields metadata from a dataclass.

    """
    fields = get_fields(cls, **filters)
    metadata = {name: field.metadata for name, field in fields.items()}
    return metadata


def get_enabled_fields(cls, name, enabled=True, default=EMPTY, **filters):
    """

    Get a dictionary of fields metadata from a dataclass filtered by enabled fields.

    """
    metadata = get_metadata(cls, **filters)

    names = []

    if isinstance(name, dict):
        name = next(iter(name.keys()))

    for key, field_meta in metadata.items():

        if default is EMPTY:
            value = field_meta[name]
        else:
            value = field_meta.get(name, default)
        if value is enabled:
            names.append(key)

    return names



"""

Additional random choice and number functions, mostly useful for data augmentation.

"""
import contextlib
import multiprocessing
import random
from functools import lru_cache
from os import getpid
from random import choices, seed

import math


@contextlib.contextmanager
def temporary_seed(seed=0):
    """

    Temporarily set random seed and revert

    """
    import random
    state = random.getstate()
    random.seed(seed)
    try:
        yield
    finally:
        random.setstate(state)


def choices_w(*data, k=1):
    """

    Allow choices and weights to be set in a more intuitive manner.

    """
    population, weights = list(zip(*data))
    return choices(population, weights, k=k)


def choice_w(*data):
    """

    Singleton version of choices_w.

    """
    return choices_w(*data, k=1)[0]


@lru_cache
def set_seed_mp_pid():
    """

    Set the random seed to be the process ID, to ensure subprocesses don't create identical randoms

    """
    if multiprocessing.parent_process():
        seed(getpid())


def prob(p=0.5) -> bool:
    """

    Return `True` with probability `p` and `False` otherwise.

    """
    return random.random() <= float(p)


def rand_log10(lower, upper) -> float|int:
    """

    Generate a random number within a specified range on a logarithmic scale

    """
    if lower == upper:
        return lower

    is_int = isinstance(lower, int) and isinstance(upper, int)

    if lower < 0 or upper < 0:
        raise ValueError('Only positives are implemented.')

    lower_shift = max(1 - lower, 0)
    upper_shift = max(1 - upper, 0) + lower_shift
    lower_shifted = lower + lower_shift
    upper_shifted = upper + upper_shift

    lower10 = math.log10(lower_shifted)
    upper10 = math.log10(upper_shifted)

    rand10 = random.uniform(lower10, upper10)
    number = 10 ** rand10
    number -= lower_shift

    if is_int:
        number = round(number)

    # TODO Add unit tests:
    # if lower10<upper10:
    #     if rand_lin < lower:
    #         raise ValueError('Return value outside bounds (too small).')
    #     if rand_lin < lower or rand_lin > upper:
    #         raise ValueError('Return value outside bounds (too large).')

    return number



import uvicorn
from dataclasses import dataclass
from fastapi import FastAPI, Request
from typing import Callable, List, Optional, Union

from fmtr.tools.environment_tools import IS_DEBUG
from fmtr.tools.iterator_tools import enlist
from fmtr.tools.logging_tools import logger


@dataclass
class Endpoint:
    """

    Endpoint-as-method config

    """
    method: Callable
    path: str
    tags: Optional[Union[str, List[str]]] = None
    method_http: Optional[Callable] = None

    def __post_init__(self):
        self.tags = enlist(self.tags)


class ApiBase:
    """

    Simple API base class, generalising endpoint-as-method config.

    """
    TITLE = 'Base API'
    HOST = '0.0.0.0'
    PORT = 8080

    def add_endpoint(self, endpoint: Endpoint):
        """

        Add endpoints from definitions using a single dataclass instance.

        """
        method_http = endpoint.method_http or self.app.post
        doc = (endpoint.method.__doc__ or '').strip() or None

        method_http(
            endpoint.path,
            tags=endpoint.tags,
            description=doc,
            summary=doc
        )(endpoint.method)

    def __init__(self):
        self.app = FastAPI(title=self.TITLE)

        for endpoint in self.get_endpoints():
            self.add_endpoint(endpoint)

        if IS_DEBUG:
            self.app.exception_handler(Exception)(self.handle_exception)

    def get_endpoints(self) -> List[Endpoint]:
        """

        Define endpoints using a dataclass instance.

        """
        endpoints = [

        ]

        return endpoints

    async def handle_exception(self, request: Request, exception: Exception):
        """

        Actually raise exceptions (e.g. for debugging) instead of returning a 500.

        """
        exception
        raise


    @classmethod
    def launch(cls):
        """

        Initialise self and launch.

        """
        self = cls()
        logger.info(f'Launching API {cls.TITLE}...')
        uvicorn.run(self.app, host=self.HOST, port=self.PORT)


if __name__ == '__main__':
    ApiBase.launch()
